#include "aoitracking.h"

#include <QProgressDialog>
#include <QtConcurrent/QtConcurrent>

#include "cv/imageprocess.h"
#include "cv/featurematching.h"
#include "opencv2/core.hpp"


AOItracking::AOItracking(DataState *state, bool interactive, int machtingAccuracy):
    machtingAccuracy(machtingAccuracy),
    interactive(interactive),
    state(state)
{
    framePairs=getFramePairs();
    for(int i=0; i<framePairs.size();i++){
        framePairsVector.append(i);
        qDebug()<<"frame pair "<<i<<" framespan" <<framePairs.at(i).first <<" "<<framePairs.at(i).second;
    }
}

int AOItracking::interactiveTracking(int breakFrame)
{
    this->breakFrame=breakFrame;
    // Create a progress dialog.
    QProgressDialog *dialog = new QProgressDialog(QString("Interactive feature based AOI tracking"), "Abort",0,0);
    dialog->setWindowModality(Qt::WindowModal);

    // Create a QFutureWatcher and connect signals and slots.
    QFutureWatcher<void> futureWatcher;
    QObject::connect(&futureWatcher, SIGNAL(finished()), dialog, SLOT(reset()));
    QObject::connect(dialog, SIGNAL(canceled()), &futureWatcher, SLOT(cancel()));

    //create wrapper
    aoiFittingWrapper wrap(this);

    // Start the computation.
    futureWatcher.setFuture(QtConcurrent::map(framePairsVector, wrap));

    // Display the dialog and start the event loop.
    dialog->exec();
    futureWatcher.waitForFinished(); //block main application
    delete dialog;

    if(breakThread==false)
        return -1;

    return this->breakFrame;
}

void AOItracking::automaticTracking()
{
    // Create a progress dialog.
    QProgressDialog *dialog = new QProgressDialog(QString("Automatic feature based AOI tracking"), "Abort",0,0);
    dialog->setWindowModality(Qt::WindowModal);

    // Create a QFutureWatcher and connect signals and slots.
    QFutureWatcher<void> futureWatcher;
    QObject::connect(&futureWatcher, SIGNAL(finished()), dialog, SLOT(reset()));
    QObject::connect(dialog, SIGNAL(canceled()), &futureWatcher, SLOT(cancel()));

    //create wrapper

    aoiFittingWrapper wrap(this);

    // Start the computation.
    futureWatcher.setFuture(QtConcurrent::map(framePairsVector, wrap));

    // Display the dialog and start the event loop.
    dialog->exec();
    futureWatcher.waitForFinished(); //block main application
    delete dialog;

}

QList<QPair<int, int> > AOItracking::getFramePairs()
{
    QList<FrameSpan*> spans = state->getCurrentObject()->getFrameSpans();
    QList<QPair<int, int> > framePairs;

    foreach (FrameSpan* span, spans) {
        if(span->type==FrameSpan::Visible){
            if(span->prevFrame!=span->nextFrame-1){
                bool append=false;
                for(int i=span->prevFrame;i<=span->nextFrame;i++){
                    if(state->getCurrentObject()->keyFrames.at(i)->polygon.isEmpty())
                        break;
                    if(i==span->nextFrame)
                        append=true;
                }
                if(append)
                    framePairs.append(QPair<int,int>(span->prevFrame,span->nextFrame));
            }
        }
    }
    if(framePairs.empty())
        framePairs.append(QPair<int,int>(-1,-1));
    return framePairs;
}

void AOItracking::aoiFitting(int &iteration)
{
    int lowerFrame=framePairs.at(iteration).first;
    int upperFrame=framePairs.at(iteration).second;
    if(lowerFrame==-1 || upperFrame ==-1)
        return;

    int delta=round((upperFrame-lowerFrame)/2);
    for(int i = lowerFrame;i<=(lowerFrame+delta) && !breakThread;i++){
        if(!transformAOI(i,i+1) && interactive && state->getCurrentObject()->keyFrames.at(i+1)->type==KeyFrame::Automatic){
            breakThread=true;
            breakFrame=i+1;
        }
    }
    for(int i = upperFrame;i>(lowerFrame+delta) && !breakThread;i--){
        if(!transformAOI(i,i-1) && interactive && state->getCurrentObject()->keyFrames.at(i-1)->type==KeyFrame::Automatic){
            breakThread=true;
            breakFrame=i-1;
        }
    }

}


bool AOItracking::transformAOI(int frame1, int frame2)
{
    if(frame1==-1|| frame2==-1)
        return false;
    ImageProcess processFrame1(state->vid->getFramePath(frame1));
    cv::Mat cropFrame1 = processFrame1.crop(state->getCurrentObject()->keyFrames.at(frame1)->polygon);
    ImageProcess processFrame2(state->vid->getFramePath(frame2));
    cv::Mat cropFrame2 = processFrame2.crop(state->getCurrentObject()->keyFrames.at(frame2)->polygon);

    FeatureMatching matching(cropFrame1,cropFrame2,state->getCurrentObject()->keyFrames.at(frame1)->polygon,state->getCurrentObject()->keyFrames.at(frame2)->polygon);
    matching.matchORB();
    matching.getGoodMatches(machtingAccuracy);

    if(matching.good_matches.size()<=7)
        return false;

    qreal x1Frame1=0,x2Frame1=0,y1Frame1=0,y2Frame1=0,x1Frame2=0,x2Frame2=0,y1Frame2=0,y2Frame2=0;
    processFrame1.getBoundingRect(state->getCurrentObject()->keyFrames.at(frame1)->polygon,x1Frame1,x2Frame1,y1Frame1,y2Frame1);
    processFrame2.getBoundingRect(state->getCurrentObject()->keyFrames.at(frame2)->polygon,x1Frame2,x2Frame2,y1Frame2,y2Frame2);

    Eigen::MatrixXd pointsFrame1=Eigen::MatrixXd::Zero(matching.good_matches.size(),2);
    Eigen::MatrixXd pointsFrame2=Eigen::MatrixXd::Zero(matching.good_matches.size(),2);
    Eigen::MatrixXd pointsDistance=Eigen::MatrixXd::Zero(matching.good_matches.size(),1);
    int matrixCount=0;
    for(std::vector<cv::DMatch>::iterator it=matching.good_matches.begin(); it !=matching.good_matches.end(); ++it){
        pointsFrame1(matrixCount/2,matrixCount%2)=(x1Frame1+matching.keyPoints1.at(it->queryIdx).pt.x);
        pointsFrame2(matrixCount/2,matrixCount%2)=(x1Frame2+matching.keyPoints2.at(it->trainIdx).pt.x);
        pointsDistance(matrixCount/2)=it->distance;

        matrixCount++;

        pointsFrame1(matrixCount/2,matrixCount%2)=(y1Frame1+matching.keyPoints1.at(it->queryIdx).pt.y);
        pointsFrame2(matrixCount/2,matrixCount%2)=(y1Frame2+matching.keyPoints2.at(it->trainIdx).pt.y);

        matrixCount++;
    }

    Eigen::Vector2d centerKeyPoints1=pointsFrame1.colwise().mean();
    Eigen::Vector2d centerKeyPoints2=pointsFrame2.colwise().mean();
    Eigen::Vector2d centerPolygon1(x2Frame1-(x2Frame1-x1Frame1)/2,y2Frame1-(y2Frame1-y1Frame1)/2);
    Eigen::Vector2d centerPolygon2(x2Frame2-(x2Frame2-x1Frame2)/2,y2Frame2-(y2Frame2-y1Frame2)/2);
    Eigen::Vector2d centerPolygon2New(x2Frame2-(x2Frame2-x1Frame2)/2,y2Frame2-(y2Frame2-y1Frame2)/2);

    centerPolygon2New=centerKeyPoints2-(centerKeyPoints1-centerPolygon1);
    Eigen::Vector2d deltaPolygon=centerPolygon2-centerPolygon2New;

    Eigen::Affine2d affineTransform;
    affineTransform=Eigen::Translation2d(x1Frame2,y1Frame2)*Eigen::Translation2d(-deltaPolygon(0),-deltaPolygon(1))*Eigen::Translation2d(-x1Frame2,-y1Frame2);

    state->getCurrentObject()->keyFrames.at(frame2)->transformPolygon(affineTransform);
    return true;
}
