#include "datastate.h"

#include <algorithm>
#include <QCoreApplication>

DataState::DataState(QObject *parent) :
    QObject(parent),
    dataStateVersion(ISEG_VERSION),
    videoLoaded(false),
    vid(NULL),
    currentFrame(0),
    currentObjectID(-1),
    colors()
{
    TrackedObject::newID = 0;

    // 22 sequentially distinguishable colors after Kelly
    colors << QColor("#222222") // black
           << QColor("#F3C300") // yellow
           << QColor("#875692") // purple
           << QColor("#F38400") // orange
           << QColor("#A1CAF1") // light blue
           << QColor("#BE0032") // red
           << QColor("#C2B280") // buff
           << QColor("#848482") // grey
           << QColor("#008856") // green
           << QColor("#E68FAC") // purplish pink
           << QColor("#0067A5") // blue
           << QColor("#F99379") // yellowish pink
           << QColor("#604E97") // violet
           << QColor("#F6A600") // orange yellow
           << QColor("#B3446C") // purplish red
           << QColor("#DCD300") // greenish yellow
           << QColor("#882D17") // reddish brown
           << QColor("#8DB600") // yellow green
           << QColor("#654522") // yellowish brown
           << QColor("#E25822") // reddish orange
           << QColor("#2B3D26"); // olive green

}

DataState::~DataState()
{
    qDeleteAll(objectsList);
    objectsList.clear();
    delete vid;
}

bool DataState::hasObjects()
{
    return !objectsList.empty();
}

void DataState::setFrame(int f)
{
    if(!videoLoaded) return;
    if(f < 0 || f >= vid->frameCount) return;
    if(f == currentFrame) return;
    currentFrame = f;
    emit hasChanged(CurrentFrameChange);
}

void DataState::nextFrame()
{
    if(currentFrame >= vid->frameCount -1) return;
    currentFrame++;
    emit hasChanged(CurrentFrameChange);
}

void DataState::prevFrame()
{
    if(currentFrame <= 0) return;
    currentFrame--;
    emit hasChanged(CurrentFrameChange);
}

void DataState::addObject()
{
    TrackedObject* obj = new TrackedObject(this);
    objectsList.append(obj);
    currentObjectID = obj->id;
    emit hasChanged(CurrentObjectChange | ObjectsListChange);
}

TrackedObject* DataState::getCurrentObject()
{
    if(currentObjectID == -1) {
        qDebug() << "No current object set!";
        return 0;
    }

    return getObjectById(currentObjectID);
}

TrackedObject *DataState::getObjectById(int id)
{
    foreach(TrackedObject* obj, objectsList) {
        if(obj->id == id) {
            return obj;
        }
    }

    qDebug() << "Could not find object with this ID";
    return 0;
}

void DataState::deleteObject(int id)
{
    bool currentChanged = false;
    bool deleted = false;
    for(int i=0; i<objectsList.length(); ++i) {
        TrackedObject* obj = objectsList[i];
        if(obj->id == id) {
            objectsList.removeAt(i);
            deleted = true;
            if(objectsList.empty()) {
                currentObjectID = -1;
                currentChanged = true;
            }
            if(id == currentObjectID) {
                int firstOrNext = 0;
                if(id == objectsList[0]->id) firstOrNext = 1;
                currentObjectID = objectsList[firstOrNext]->id;
                currentChanged = true;
            }
            break;
        }
    }
    if(!deleted) {
        qWarning() << "Could not find object with this ID";
    } else if(currentChanged) {
        emit hasChanged(ObjectsListChange | CurrentObjectChange);
    } else {
        emit hasChanged(ObjectsListChange);
    }

}

QList<TrackedObject *> DataState::getZOrderSortedList(int frame)
{
    QList<TrackedObject *> sorted(this->objectsList);

    std::sort(sorted.begin(), sorted.end(), TrackedObject::combareByZOrderOfFrame(frame));

    return sorted;
}

void DataState::setCurrentObject(int id)
{
    currentObjectID = id;
    emit hasChanged(CurrentObjectChange);
}
